import os
import datetime
import pandas as pd
import numpy as np
from ab_test_advanced_toolkit.analyzer import ABTestAnalyzer
from typing import Tuple

import logging

logger = logging.getLogger(__name__)

# Create folder 'data/{now_timestamp_utc}' if it does not exist
folder_name = os.path.join(
    'data', 
    datetime.datetime.now(datetime.timezone.utc).strftime('%Y-%m-%d_%H-%M-%S')
)
os.makedirs(folder_name, exist_ok=True)
def describe_dataset(df):
    logging.debug("First 5 rows of the dataset:")
    logging.debug(df.head())
    
    logging.debug("\nStatistical summary of the dataset:")
    logging.debug(df.describe())
    
    logging.debug("\nCorrelation between 'value' and 'pre_test_value':")
    logging.debug(df[['value', 'pre_test_value']].corr())
    
    logging.debug("\nAverage 'value' by A/B groups:")
    ab_means = df.groupby('abgroup')['value'].mean()
    logging.debug(ab_means)

def generate_synthetic_data(num_users, countries, platforms, user_segments, ab_groups, base_increase_percentage, noise_level=1.0, correlation_level=0.5, seed=40):
    # Set seed for reproducibility
    np.random.seed(seed)
    
    # Generate synthetic data
    data = {
        'userid': range(1, num_users + 1),
        'country': np.random.choice(countries, num_users),
        'platform': np.random.choice(platforms, num_users),
        'user_segment': np.random.choice(user_segments, num_users),
        'abgroup': np.random.choice(ab_groups, num_users, p=[1/3, 1/3, 1/3]),
        'age': np.random.randint(18, 65, num_users),  # Age between 18 and 65
        'engagement_score': np.random.rand(num_users) * 10,  # Random score between 0 and 10
    }
    df = pd.DataFrame(data)

    # Calculate the base effect from features
    base_value = 10 + df['age'] / 10 + df['engagement_score']

    # Pre-compute indexes for countries, platforms, and user_segments to avoid using apply
    df['country_idx'] = df['country'].apply(lambda x: countries.index(x) % 3)
    df['platform_idx'] = df['platform'].apply(lambda x: platforms.index(x) % 2)
    df['segment_idx'] = df['user_segment'].apply(lambda x: user_segments.index(x) % 4)

    # Vectorized operations for country, platform, and segment effects
    country_effect = np.random.uniform(-1, 5, num_users) * df['country_idx']
    platform_effect = np.random.uniform(-1, 5, num_users) * df['platform_idx']
    segment_effect = np.random.uniform(-1, 5, num_users) * df['segment_idx']

    category_effect = country_effect + platform_effect + segment_effect

    # Group effect with added randomness
    group_effect = np.where(
        df['abgroup'] == 'b', 
        np.random.uniform(-1 * base_increase_percentage, base_increase_percentage * 3, num_users), 
        np.random.uniform(-2 * base_increase_percentage, base_increase_percentage * 2, num_users),
    )

    # Define mean and covariance for multivariate normal distribution
    mean = [0, 0]
    cov = [[1, correlation_level], [correlation_level, 1]]
    
    # Generate correlated values for 'value' and 'pre_test_value'
    correlated_values = np.random.multivariate_normal(mean, cov, num_users)

    # Calculate the noise
    noise = np.random.normal(0, noise_level, num_users)

    # Calculate the final value with added nonlinear category effect and noise
    df['value'] = base_value * (1 + category_effect) * (1 + group_effect) + correlated_values[:, 0] + noise

    # Generating pre-test value with a slightly different formula to introduce non-linearity and noise
    nonlinear_component = 0.5 * np.sin(df['engagement_score'] / 2) * np.random.uniform(-1, 1, num_users)
    df['pre_test_value'] = base_value * (1 + category_effect) * (1 + correlated_values[:, 1]) + nonlinear_component + noise


    describe_dataset(df)

    return df

def create_dataframes(df: pd.DataFrame) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]:
    event_data1 = pd.DataFrame({
        "timestamp": pd.to_datetime(["2022-12-10"] * len(df)),
        "userid": df["userid"],
        "event_name": ["purchase"] * len(df),
        "purchase_value": df["value"]
    })

    event_data2 = pd.DataFrame({
        "timestamp": pd.to_datetime(["2022-12-01"] * len(df)),
        "userid": df["userid"],
        "event_name": ["purchase"] * len(df),
        "purchase_value": df["pre_test_value"]
    })

    user_allocations = pd.DataFrame({
        "timestamp": pd.to_datetime(["2022-12-05"] * len(df)),
        "userid": df["userid"],
        "abgroup": df["abgroup"]
    })

    user_properties = pd.DataFrame({
        "userid": df["userid"],
        "age": df["age"],
        "country": df["country"],
        "device_type": df["platform"],
        "membership_status": ["Free"] * len(df)
    })

    event_data = pd.concat([event_data1, event_data2])
    event_data['timestamp'] = pd.to_datetime(event_data['timestamp'])

    return event_data, user_allocations, user_properties

def run_analysis(df: pd.DataFrame):
    event_data, user_allocations, user_properties = create_dataframes(df)
    
    # Analysis without enhancement
    analyzer_no_enhancement = ABTestAnalyzer(event_data, user_allocations, "a1", user_properties, mode="no_enhancement")
    results_no_enhancement = analyzer_no_enhancement.calculate_event_attribute_sum_per_user('purchase', 'purchase_value')

    # Analysis with CUPED
    analyzer_cuped = ABTestAnalyzer(event_data, user_allocations, "a1", user_properties, mode="cuped")
    results_cuped = analyzer_cuped.calculate_event_attribute_sum_per_user('purchase', 'purchase_value')

    # Analysis with gboost CUPED
    analyzer_gboost_cuped = ABTestAnalyzer(event_data, user_allocations, "a1", user_properties, mode="gboost_cuped")
    results_gboost_cuped = analyzer_gboost_cuped.calculate_event_attribute_sum_per_user('purchase', 'purchase_value')

    return {
        "no_enhancement": results_no_enhancement,
        "cuped": results_cuped,
        "gboost_cuped": results_gboost_cuped
    }

if __name__ == "__main__":
    # Set up logging
    logger.setLevel(logging.DEBUG)

    # Generate synthetic data
    generated_data = generate_synthetic_data(
        num_users=100000,
        countries=['US', 'UK', 'DE', 'FR', 'CA', 'AU', 'JP', 'IN'],
        platforms=['iOS', 'Android', 'Web', 'Desktop'],
        user_segments=['Segment_1', 'Segment_2', 'Segment_3', 'Segment_4'],
        ab_groups=['a1', 'a2', 'b'],
        noise_level=1.0,
        base_increase_percentage=0.05,
        correlation_level=0.5,
    )

    # Save the data to a CSV file
    file_name = '0_generated_data.csv'
    full_path = os.path.join(folder_name, file_name)
    generated_data.to_csv(full_path, index=False)
    logging.info(f"Data saved to {full_path}")

    # Run analysis on the generated data
    results = run_analysis(generated_data)
    logging.info("Analysis results (no enhancement): %s", results["no_enhancement"])
    logging.info("Analysis results (cuped): %s", results["cuped"])
    logging.info("Analysis results (gboost_cuped): %s", results["gboost_cuped"])




from typing import Tuple, List, Optional

import pandas as pd

from ab_test_advanced_toolkit.data_validation import validate_data
from ab_test_advanced_toolkit.metrics import Metric, MetricType, MetricParams, MetricResult, AggregationOperation
from ab_test_advanced_toolkit.stat_significance import StatTests
from ab_test_advanced_toolkit.vizualizer import format_metrics_to_html


import logging

# Function to set up logging configuration
def setup_logging(level=logging.INFO):
    logging.basicConfig(level=level, format='%(asctime)s - %(levelname)s - %(message)s')
    logger = logging.getLogger(__name__)
    return logger

class ABTestAnalyzer:
    def __init__(self, event_data: pd.DataFrame, ab_test_allocations: pd.DataFrame, control_group_name: str,
                 user_properties: Optional[pd.DataFrame] = None, mode="gboost_cuped", logging_level=logging.INFO):
        """
        Initializes the ABTestAnalyzer with event data, AB test allocations, control group name, user properties, and mode.
        :param event_data: DataFrame containing event data. Expected pandas format: |timestamp|userid|event_name|attribute_1|attribute_2|...
        :param ab_test_allocations: DataFrame containing AB test allocations. Expected pandas format: |timestamp|userid|abgroup|
        :param control_group_name: The name of the control group.
        :param user_properties: DataFrame containing user properties. Expected pandas format: |userid|property_1|property_2|...
        :param mode: Mode of enhancement ("no_enhancement", "cuped", "gboost_cuped").
        :param logging_level: The logging level to be used (e.g., logging.INFO, logging.DEBUG).
        """

        self.logger = setup_logging(logging_level)

        assert mode in ["no_enhancement", "cuped", "gboost_cuped"], "Invalid mode"

        self.control_group_name = control_group_name
        self.test_group_names = [group for group in ab_test_allocations['abgroup'].unique() if
                                 group != control_group_name]

        validate_data(event_data, ab_test_allocations, control_group_name)

        self.ab_test_allocations = ab_test_allocations.set_index("userid")
        self.event_data = event_data
        self.user_properties = user_properties
        self.calculated_metrics: List[Metric] = []
        self.mode = mode

    @staticmethod
    def _merge_and_aggregate(event_data: pd.DataFrame, ab_test_allocations: pd.DataFrame, event_name: str,
                             operation: AggregationOperation,
                             attribute_name=None,
                             pretest=False) -> Tuple[pd.DataFrame, pd.DataFrame]:
        """
        Merges event data with AB test allocations and aggregates data based on specified attributes and operation.
        Adds support for conversion operation.

        :param event_data: DataFrame containing event data. Expected pandas format: |timestamp|userid|event_name|attribute_1|attribute_2|...
        :param ab_test_allocations: DataFrame containing AB test allocations. Expected pandas format: |timestamp|userid|abgroup|
        :param event_name: The name of the event to filter on.
        :param attribute_name: The name of the attribute to aggregate sum
        :param operation: The aggregation operation ('sum', 'count', or 'conversion').
        :param pretest: Boolean indicating whether to process pretest (True) or intest (False) data.
        :return: raw merged data, metrics data
        """
        filtered_event_data = event_data[event_data["event_name"] == event_name]

        event_data_with_alloc = pd.merge(filtered_event_data, ab_test_allocations[['timestamp']], left_on='userid',
                                         right_index=True, how='left', suffixes=('_event', '_alloc'))

        # Filter for pretest or intest events
        if pretest:
            filtered_events = event_data_with_alloc[
                event_data_with_alloc['timestamp_event'] < event_data_with_alloc['timestamp_alloc']]
        else:
            filtered_events = event_data_with_alloc[
                event_data_with_alloc['timestamp_event'] >= event_data_with_alloc['timestamp_alloc']]

        # Apply aggregation operation
        if operation == AggregationOperation.CONVERSION:
            # Mark each user as 1 (converted) if the event occurred, else 0
            aggregated_data = filtered_events.groupby("userid").size().gt(0).astype(int).to_frame(
                name='conversion_status')
        elif operation == AggregationOperation.COUNT:
            aggregated_data = filtered_events.groupby("userid").agg({"event_name": "count"})
        elif operation == AggregationOperation.SUM:
            aggregated_data = filtered_events.groupby("userid").agg({attribute_name: "sum"})
        else:
            raise ValueError(f"Unsupported aggregation operation: {operation}")

        # Merge aggregated data with AB test allocations and fill missing values with 0
        merged_data = pd.merge(ab_test_allocations[['abgroup']], aggregated_data, left_index=True, right_index=True,
                               how="left").fillna(0)
        result = merged_data.groupby("abgroup").mean()

        return merged_data, result

    def calculate_event_count_per_user(self, event_name: str) -> pd.DataFrame:
        """
        Calculates the count of events per user for the specified event name.
        :param event_name: The name of the event to calculate the count for.
        :return: DataFrame with the calculated event count per user per group
        """
        merged_pretest, _ = self._merge_and_aggregate(self.event_data, self.ab_test_allocations, event_name,
                                                      AggregationOperation.COUNT, pretest=True)

        merged_intest, result_intest = self._merge_and_aggregate(self.event_data, self.ab_test_allocations, event_name,
                                                                 AggregationOperation.COUNT)

        if self.mode == "gboost_cuped":
            stat_test = StatTests.calculate_gboost_cuped_and_compare(merged_pretest, merged_intest, self.user_properties,
                                                                self.control_group_name, self.test_group_names, True)
        elif self.mode == "cuped":
            stat_test = StatTests.calculate_cuped_and_compare(merged_pretest, merged_intest,
                                                                self.control_group_name, self.test_group_names)
        else:
            stat_test = StatTests.calculate_gboost_cuped_and_compare(merged_pretest, merged_intest,
                                                                    self.user_properties,
                                                                    self.control_group_name, self.test_group_names,
                                                                    False)

        metric_output = Metric(MetricType.EVENT_COUNT_PER_USER, MetricParams(event_name),
                                              MetricResult(result_intest, self.control_group_name,
                                                           self.test_group_names, stat_test))
        self.calculated_metrics.append(metric_output)
        return metric_output

    def calculate_event_attribute_sum_per_user(self, event_name: str, attribute_name: str) -> pd.DataFrame:
        """
        Calculates the sum of a specified attribute per user for the specified event name.
        :param event_name: The event name to calculate the attribute sum for.
        :param attribute_name:  The name of the attribute to calculate the sum for.
        :return: DataFrame with the calculated attribute sum per user per group
        """

        if attribute_name not in self.event_data.columns:
            raise ValueError(f"Attribute {attribute_name} not found in event data.")
        if self.event_data[attribute_name].dtype not in [int, float]:
            raise ValueError(f"Attribute {attribute_name} is not numeric.")

        merged_pretest, _ = self._merge_and_aggregate(self.event_data, self.ab_test_allocations, event_name,
                                                    AggregationOperation.SUM, attribute_name, pretest=True)

        merged_intest, result_intest = self._merge_and_aggregate(self.event_data, self.ab_test_allocations, event_name,
                                                                AggregationOperation.SUM, attribute_name)

        if self.mode == "gboost_cuped":
            stat_test = StatTests.calculate_gboost_cuped_and_compare(merged_pretest, merged_intest, self.user_properties,
                                                                self.control_group_name, self.test_group_names, True)
        elif self.mode == "cuped":
            stat_test = StatTests.calculate_cuped_and_compare(merged_pretest, merged_intest,
                                                                self.control_group_name, self.test_group_names)
        else:
            stat_test = StatTests.calculate_gboost_cuped_and_compare(merged_pretest, merged_intest,
                                                                    self.user_properties,
                                                                    self.control_group_name, self.test_group_names,
                                                                    False)

        metric_output = Metric(MetricType.EVENT_ATTRIBUTE_SUM_PER_USER, MetricParams(event_name, attribute_name),
                MetricResult(result_intest, self.control_group_name, self.test_group_names, stat_test))
        self.calculated_metrics.append(metric_output)
        return metric_output

    def calculate_conversion(self, target_event: str) -> pd.DataFrame:
        """
        Calculates the conversion rate for the specified target event.
        :param target_event: The name of the event to calculate the conversion rate for.
        :return: DataFrame with the calculated conversion rate per user per group
        """
        merged_intest, result_intest = self._merge_and_aggregate(self.event_data, self.ab_test_allocations,
                                                                 target_event,
                                                                 AggregationOperation.CONVERSION)
        stat_test = StatTests.calculate_t_test_for_dataset(merged_intest, self.control_group_name,
                                                           self.test_group_names)

        metric_output = Metric(
            MetricType.CONVERSION_RATE,
            MetricParams(target_event),
            MetricResult(result_intest, self.control_group_name, self.test_group_names, stat_test))
        self.calculated_metrics.append(metric_output)
        return metric_output

    def save_report(self, filename: str):
        res = format_metrics_to_html(self.calculated_metrics, self.control_group_name, self.test_group_names)
        with open(filename, 'w') as f:
            f.write(res)
